/*
 * Copyright © 2022 Soren Stoutner <soren@stoutner.com>.
 *
 * This file is part of Privacy Browser PC <https://www.stoutner.com/privacy-browser-pc>.
 *
 * Privacy Browser PC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Privacy Browser PC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Privacy Browser PC.  If not, see <http://www.gnu.org/licenses/>.
 */

// Application headers.
#include "BrowserWindow.h"
#include "Settings.h"
#include "ui_SettingsPrivacy.h"
#include "ui_SettingsGeneral.h"
#include "dialogs/DomainSettingsDialog.h"
#include "helpers/SearchEngineHelper.h"
#include "helpers/UserAgentHelper.h"

// KDE Frameworks headers.
#include <KActionCollection>
#include <KToolBar>

// Qt toolkit headers.
#include <QInputDialog>
#include <QStatusBar>

BrowserWindow::BrowserWindow() : KXmlGuiWindow()
{
    // Instantiate the main view pointer.
    browserViewPointer = new BrowserView(this);

    // Set the main view as the central widget.
    setCentralWidget(browserViewPointer);

    // Get a handle for the action collection.
    KActionCollection *actionCollectionPointer = this->actionCollection();

    // Add the standard actions.
    QAction *backActionPointer = KStandardAction::back(this, SLOT(back()), actionCollectionPointer);
    QAction *forwardActionPointer = KStandardAction::forward(this, SLOT(forward()), actionCollectionPointer);
    KStandardAction::home(this, SLOT(home()), actionCollectionPointer);
    KStandardAction::openNew(this, SLOT(fileNew()), actionCollectionPointer);
    KStandardAction::quit(qApp, SLOT(closeAllWindows()), actionCollectionPointer);
    KStandardAction::preferences(this, SLOT(settingsConfigure()), actionCollectionPointer);
    KStandardAction::redisplay(this, SLOT(refresh()), actionCollectionPointer);

    // Add the custom actions.
    userAgentPrivacyBrowserActionPointer = actionCollectionPointer->addAction(QStringLiteral("user_agent_privacy_browser"));
    userAgentFirefoxLinuxActionPointer = actionCollectionPointer->addAction(QStringLiteral("user_agent_firefox_linux"));
    userAgentChromiumLinuxActionPointer = actionCollectionPointer->addAction(QStringLiteral("user_agent_chromium_linux"));
    userAgentFirefoxWindowsActionPointer = actionCollectionPointer->addAction(QStringLiteral("user_agent_firefox_windows"));
    userAgentChromeWindowsActionPointer = actionCollectionPointer->addAction(QStringLiteral("user_agent_chrome_windows"));
    userAgentEdgeWindowsActionPointer = actionCollectionPointer->addAction(QStringLiteral("user_agent_edge_windows"));
    userAgentSafariMacosActionPointer = actionCollectionPointer->addAction(QStringLiteral("user_agent_safari_macos"));
    userAgentCustomActionPointer = actionCollectionPointer->addAction(QStringLiteral("user_agent_custom"));
    zoomFactorActionPointer = actionCollectionPointer->addAction(QStringLiteral("zoom_factor"));
    searchEngineMojeekActionPointer = actionCollectionPointer->addAction(QStringLiteral("search_engine_mojeek"));
    searchEngineMonoclesActionPointer = actionCollectionPointer->addAction(QStringLiteral("search_engine_monocles"));
    searchEngineMetagerActionPointer = actionCollectionPointer->addAction(QStringLiteral("search_engine_metager"));
    searchEngineGoogleActionPointer = actionCollectionPointer->addAction(QStringLiteral("search_engine_google"));
    searchEngineBingActionPointer = actionCollectionPointer->addAction(QStringLiteral("search_engine_bing"));
    searchEngineYahooActionPointer = actionCollectionPointer->addAction(QStringLiteral("search_engine_yahoo"));
    searchEngineCustomActionPointer = actionCollectionPointer->addAction(QStringLiteral("search_engine_custom"));
    javaScriptActionPointer = actionCollectionPointer->addAction(QStringLiteral("javascript"));
    domainSettingsActionPointer = actionCollectionPointer->addAction(QStringLiteral("domain_settings"));

    // Create the action groups
    QActionGroup *userAgentActionGroupPointer = new QActionGroup(this);
    QActionGroup *searchEngineActionGroupPointer = new QActionGroup(this);

    // Add the actions to the groups.
    userAgentActionGroupPointer->addAction(userAgentPrivacyBrowserActionPointer);
    userAgentActionGroupPointer->addAction(userAgentFirefoxLinuxActionPointer);
    userAgentActionGroupPointer->addAction(userAgentChromiumLinuxActionPointer);
    userAgentActionGroupPointer->addAction(userAgentFirefoxWindowsActionPointer);
    userAgentActionGroupPointer->addAction(userAgentChromeWindowsActionPointer);
    userAgentActionGroupPointer->addAction(userAgentEdgeWindowsActionPointer);
    userAgentActionGroupPointer->addAction(userAgentSafariMacosActionPointer);
    userAgentActionGroupPointer->addAction(userAgentCustomActionPointer);
    searchEngineActionGroupPointer->addAction(searchEngineMojeekActionPointer);
    searchEngineActionGroupPointer->addAction(searchEngineMonoclesActionPointer);
    searchEngineActionGroupPointer->addAction(searchEngineMetagerActionPointer);
    searchEngineActionGroupPointer->addAction(searchEngineGoogleActionPointer);
    searchEngineActionGroupPointer->addAction(searchEngineBingActionPointer);
    searchEngineActionGroupPointer->addAction(searchEngineYahooActionPointer);
    searchEngineActionGroupPointer->addAction(searchEngineCustomActionPointer);

    // Set some actions to be checkable.
    userAgentPrivacyBrowserActionPointer->setCheckable(true);
    userAgentFirefoxLinuxActionPointer->setCheckable(true);
    userAgentChromiumLinuxActionPointer->setCheckable(true);
    userAgentFirefoxWindowsActionPointer->setCheckable(true);
    userAgentChromeWindowsActionPointer->setCheckable(true);
    userAgentEdgeWindowsActionPointer->setCheckable(true);
    userAgentSafariMacosActionPointer->setCheckable(true);
    userAgentCustomActionPointer->setCheckable(true);
    searchEngineMojeekActionPointer->setCheckable(true);
    searchEngineMonoclesActionPointer->setCheckable(true);
    searchEngineMetagerActionPointer->setCheckable(true);
    searchEngineGoogleActionPointer->setCheckable(true);
    searchEngineBingActionPointer->setCheckable(true);
    searchEngineYahooActionPointer->setCheckable(true);
    searchEngineCustomActionPointer->setCheckable(true);

    // Set the non-mutable action text.
    userAgentPrivacyBrowserActionPointer->setText(UserAgentHelper::PRIVACY_BROWSER_TRANSLATED);
    userAgentFirefoxLinuxActionPointer->setText(UserAgentHelper::FIREFOX_LINUX_TRANSLATED);
    userAgentChromiumLinuxActionPointer->setText(UserAgentHelper::CHROMIUM_LINUX_TRANSLATED);
    userAgentFirefoxWindowsActionPointer->setText(UserAgentHelper::FIREFOX_WINDOWS_TRANSLATED);
    userAgentChromeWindowsActionPointer->setText(UserAgentHelper::CHROME_WINDOWS_TRANSLATED);
    userAgentEdgeWindowsActionPointer->setText(UserAgentHelper::EDGE_WINDOWS_TRANSLATED);
    userAgentSafariMacosActionPointer->setText(UserAgentHelper::SAFARI_MACOS_TRANSLATED);
    searchEngineMojeekActionPointer->setText(i18nc("Search engine", "Mojeek"));
    searchEngineMonoclesActionPointer->setText(i18nc("Search engine", "Monocles"));
    searchEngineMetagerActionPointer->setText(i18nc("Search engine", "MetaGer"));
    searchEngineGoogleActionPointer->setText(i18nc("Search engine", "Google"));
    searchEngineBingActionPointer->setText(i18nc("Search engine", "Bing"));
    searchEngineYahooActionPointer->setText(i18nc("Search engine", "Yahoo"));
    javaScriptActionPointer->setText(i18nc("JavaScript button", "JavaScript"));
    domainSettingsActionPointer->setText(i18nc("Domain Settings button", "Domain Settings"));

    // Set the action icons.
    userAgentPrivacyBrowserActionPointer->setIcon(QIcon(":/icons/privacy-mode"));
    userAgentFirefoxLinuxActionPointer->setIcon(QIcon::fromTheme(QStringLiteral("firefox-esr")));
    userAgentChromiumLinuxActionPointer->setIcon(QIcon::fromTheme(QStringLiteral("chromium")));
    userAgentFirefoxWindowsActionPointer->setIcon(QIcon::fromTheme(QStringLiteral("firefox-esr")));
    userAgentChromeWindowsActionPointer->setIcon(QIcon::fromTheme(QStringLiteral("chromium")));
    userAgentEdgeWindowsActionPointer->setIcon(QIcon::fromTheme(QStringLiteral("user-group-properties")));
    userAgentSafariMacosActionPointer->setIcon(QIcon::fromTheme(QStringLiteral("user-group-properties")));
    userAgentCustomActionPointer->setIcon(QIcon::fromTheme(QStringLiteral("user-group-properties")));
    searchEngineMojeekActionPointer->setIcon(QIcon::fromTheme(QStringLiteral("search")));
    searchEngineMonoclesActionPointer->setIcon(QIcon::fromTheme(QStringLiteral("search")));
    searchEngineMetagerActionPointer->setIcon(QIcon::fromTheme(QStringLiteral("search")));
    searchEngineGoogleActionPointer->setIcon(QIcon::fromTheme(QStringLiteral("im-google")));
    searchEngineBingActionPointer->setIcon(QIcon::fromTheme(QStringLiteral("search")));
    searchEngineYahooActionPointer->setIcon(QIcon::fromTheme(QStringLiteral("im-yahoo")));
    searchEngineCustomActionPointer->setIcon(QIcon::fromTheme(QStringLiteral("search")));
    zoomFactorActionPointer->setIcon(QIcon::fromTheme(QStringLiteral("zoom")));
    domainSettingsActionPointer->setIcon(QIcon::fromTheme(QStringLiteral("network-server-symbolic")));

    // Update the on-the-fly menus.
    connect(browserViewPointer, SIGNAL(updateUserAgentActions(QString)), this, SLOT(updateUserAgentActions(QString)));
    connect(browserViewPointer, SIGNAL(updateZoomFactorAction(double)), this, SLOT(updateZoomFactorAction(double)));
    connect(browserViewPointer, SIGNAL(updateSearchEngineActions(QString)), this, SLOT(updateSearchEngineActions(QString)));

    // Apply the on-the-fly settings when selected.
    connect(userAgentActionGroupPointer, SIGNAL(triggered(QAction*)), browserViewPointer, SLOT(applyOnTheFlyUserAgent(QAction*)));
    connect(searchEngineActionGroupPointer, SIGNAL(triggered(QAction*)), browserViewPointer, SLOT(applyOnTheFlySearchEngine(QAction*)));

    // Display dialogs.
    connect(zoomFactorActionPointer, SIGNAL(triggered()), this, SLOT(getZoomFactorFromUser()));

    // Connect the URL toolbar actions.
    connect(javaScriptActionPointer, SIGNAL(triggered()), this, SLOT(toggleJavaScript()));
    connect(domainSettingsActionPointer, SIGNAL(triggered()), this, SLOT(openDomainSettings()));

    // Update the URL toolbar actions.
    connect(browserViewPointer, SIGNAL(updateBackAction(bool)), backActionPointer, SLOT(setEnabled(bool)));
    connect(browserViewPointer, SIGNAL(updateForwardAction(bool)), forwardActionPointer, SLOT(setEnabled(bool)));
    connect(browserViewPointer, SIGNAL(updateJavaScriptAction(bool)), this, SLOT(updateJavaScriptAction(bool)));

    // Setup the GUI based on the browser_ui.rc file.
    setupGUI(StandardWindowOption::Default, ("browser_ui.rc"));

    // Get a handle for the URL toolbar.
    KToolBar *urlToolBarPointer = toolBar(QStringLiteral("url_toolbar"));

    // Create a URL line edit.
    urlLineEditPointer = new KLineEdit();

    // Populate the URL toolbar.
    urlToolBarPointer->addWidget(urlLineEditPointer);
    urlToolBarPointer->addAction(javaScriptActionPointer);
    urlToolBarPointer->addAction(domainSettingsActionPointer);

    // Load a new URL from the URL line edit.
    connect(urlLineEditPointer, SIGNAL(returnKeyPressed(const QString)), this, SLOT(loadUrlFromLineEdit(const QString)));

    // Update the URL line edit on page loads.
    connect(browserViewPointer, SIGNAL(updateUrlLineEdit(QString)), this, SLOT(updateUrlLineEdit(QString)));

    // Get a handle for the status bar.
    QStatusBar *statusBarPointer = statusBar();

    // Create a progress bar.
    progressBarPointer = new QProgressBar();

    // Add the progress bar to to the status bar.
    statusBarPointer->addPermanentWidget(progressBarPointer);

    // Update the status bar with the URL when a link is hovered.
    connect(browserViewPointer, SIGNAL(linkHovered(QString)), statusBarPointer, SLOT(showMessage(QString)));

    // Update the progress bar.
    connect(browserViewPointer, SIGNAL(showProgressBar(const int)), this, SLOT(showProgressBar(const int)));
    connect(browserViewPointer, SIGNAL(hideProgressBar()), progressBarPointer, SLOT(hide()));

    // Get the URL line edit palettes.
    noDomainSettingsPalette = urlLineEditPointer->palette();
    domainSettingsPalette = urlLineEditPointer->palette();

    // Modify the domain settings palette.
    domainSettingsPalette.setColor(QPalette::Base, QColor("#C8E6C9"));

    // Update the applied palette.
    connect(browserViewPointer, SIGNAL(updateDomainSettingsIndicator(bool)), this, SLOT(updateDomainSettingsIndicator(bool)));

    // Load the initial website.
    browserViewPointer->loadInitialWebsite();
}

void BrowserWindow::back() const
{
    // Remove the focus from the URL line edit.
    urlLineEditPointer->clearFocus();

    // Go back.
    browserViewPointer->back();
}

void BrowserWindow::fileNew() const
{
    // Display a new instance of Privacy Browser.
    (new BrowserWindow)->show();
}

void BrowserWindow::forward() const
{
    // Remove the focus from the URL line edit.
    urlLineEditPointer->clearFocus();

    // Go forward.
    browserViewPointer->forward();
}

void BrowserWindow::getZoomFactorFromUser()
{
    // Create an OK flag.
    bool okClicked;

    // Display a dialog to get the new zoom factor from the user.  Format the double to display two decimals and have a 0.25 step.
    double newZoomFactor = QInputDialog::getDouble(this, i18nc("The on-the-fly zoom factor dialog title", "On-The-Fly Zoom Factor"),
                                                   i18nc("The instruction text of the on-the-fly zoom factor dialog", "Enter a zoom factor between 0.25 and 5.00"),
                                                   currentZoomFactor, .025, 5.00, 2, &okClicked, Qt::WindowFlags(), 0.25);

    // Update the zoom factor if the user clicked OK.
    if (okClicked)
    {
        // Update the current zoom factor.
        currentZoomFactor = newZoomFactor;

        // Set the new zoom factor.
        browserViewPointer->applyOnTheFlyZoomFactor(newZoomFactor);

        // Update the on-the-fly action text.
        updateZoomFactorAction(newZoomFactor);
    }
}

void BrowserWindow::home() const
{
    // Remove the focus from the URL line edit.
    urlLineEditPointer->clearFocus();

    // Go home.
    browserViewPointer->home();
}

void BrowserWindow::loadUrlFromLineEdit(const QString &url) const
{
    // Remove the focus from the URL line edit.
    urlLineEditPointer->clearFocus();

    // Load the URL.
    browserViewPointer->loadUrlFromLineEdit(url);
}

void BrowserWindow::openDomainSettings() const
{
    // Remove the focus from the URL line edit.
    urlLineEditPointer->clearFocus();

    // Instantiate the domain settings window.
    DomainSettingsDialog *domainSettingsDialogPointer = new DomainSettingsDialog();

    // Set the dialog window title.
    domainSettingsDialogPointer->setWindowTitle(i18nc("The domain settings dialog title", "Domain Settings"));

    // Set the modality.
    domainSettingsDialogPointer->setWindowModality(Qt::WindowModality::WindowModal);;

    // Show the dialog.
    domainSettingsDialogPointer->show();

    // Reload the tabs when domain settings are updated.
    connect(domainSettingsDialogPointer, SIGNAL(domainSettingsUpdated()), browserViewPointer, SLOT(applyDomainSettingsAndReload()));
}

void BrowserWindow::refresh() const
{
    // Remove the focus from the URL line edit.
    urlLineEditPointer->clearFocus();

    // Refresh the web page.
    browserViewPointer->refresh();
}

void BrowserWindow::showProgressBar(const int &progress) const
{
    // Set the progress bar value.
    progressBarPointer->setValue(progress);

    // Show the progress bar.
    progressBarPointer->show();
}

void BrowserWindow::toggleJavaScript() const
{
    // Remove the focus from the URL line edit.
    urlLineEditPointer->clearFocus();

    // Toggle JavaScript.
    browserViewPointer->toggleJavaScript();
}

QSize BrowserWindow::sizeHint() const
{
    // Return the default window size.
    return QSize(1500, 1200);
}

void BrowserWindow::settingsConfigure()
{
    // Check to make sure the dialog box isn't already displayed.
    if (KConfigDialog::exists(QStringLiteral("settings")))
    {
        // Show the existing config dialog if it is hidden.
        configDialogPointer->show();

        // Raise the existing config dialog if it is below other windows.
        configDialogPointer->raise();

        // Restore the existing config dialog if it has been minimized.
        if (configDialogPointer->isMinimized()) {
            configDialogPointer->showNormal();
        }

        // Activate the existing config dialog, which brings its virtual desktop into focus.
        configDialogPointer->activateWindow();
    }
    else
    {
        // Create the settings widgets.
        QWidget *privacySettingsWidgetPointer = new QWidget;
        QWidget *generalSettingsWidgetPointer = new QWidget;

        // Instantiate the settings UI.
        Ui::PrivacySettings privacySettingsUi;
        Ui::GeneralSettings generalSettingsUi;

        // Setup the UI to display the settings widgets.
        privacySettingsUi.setupUi(privacySettingsWidgetPointer);
        generalSettingsUi.setupUi(generalSettingsWidgetPointer);

        // Get handles for the widgets.
        QComboBox *userAgentComboBoxPointer = privacySettingsUi.kcfg_userAgent;
        userAgentLabelPointer = privacySettingsUi.userAgentLabel;
        QComboBox *searchEngineComboBoxPointer = generalSettingsUi.kcfg_searchEngine;
        searchEngineLabelPointer = generalSettingsUi.searchEngineLabel;

        // Populate the combo box labels.
        updateUserAgentLabel(userAgentComboBoxPointer->currentText());
        updateSearchEngineLabel(searchEngineComboBoxPointer->currentText());

        // Update the labels when the combo boxes change.
        connect(userAgentComboBoxPointer, SIGNAL(currentTextChanged(const QString)), this, SLOT(updateUserAgentLabel(const QString)));
        connect(searchEngineComboBoxPointer, SIGNAL(currentTextChanged(const QString)), this, SLOT(updateSearchEngineLabel(const QString)));

        // Instantiate a settings config dialog from the settings.kcfg file.
        configDialogPointer = new KConfigDialog(this, QStringLiteral("settings"), Settings::self());

        // Add the settings widgets as config dialog pages.
        configDialogPointer->addPage(privacySettingsWidgetPointer, i18nc("@title:tab", "Privacy"), QStringLiteral("privacy-browser"));
        configDialogPointer->addPage(generalSettingsWidgetPointer, i18nc("@title:tab", "General"), QStringLiteral("breeze-settings"));

        // Delete the config dialog when it is closed.
        configDialogPointer->setAttribute(Qt::WA_DeleteOnClose);

        // Make it so.
        configDialogPointer->show();

        // TODO.  KConfigDialog does not respect expanding size policies.  <https://redmine.stoutner.com/issues/823>
        //configDialogPointer->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
        //privacySettingsWidgetPointer->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
        //generalSettingsWidgetPointer->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
        //configDialogPointer->adjustSize();

        // Expand the config dialog.
        configDialogPointer->resize(1000, 500);

        // Apply the settings when they are updated.
        connect(configDialogPointer, SIGNAL(settingsChanged(QString)), browserViewPointer, SLOT(applyApplicationSettings()));
        connect(configDialogPointer, SIGNAL(settingsChanged(QString)), browserViewPointer, SLOT(applyDomainSettingsAndReload()));
    }
}

void BrowserWindow::updateDomainSettingsIndicator(const bool &status) const
{
    // Set the domain palette according to the status.
    if (status) urlLineEditPointer->setPalette(domainSettingsPalette);
    else urlLineEditPointer->setPalette(noDomainSettingsPalette);
}

void BrowserWindow::updateJavaScriptAction(const bool &isEnabled) const
{
    // Set the icon according to the status.
    if (isEnabled) javaScriptActionPointer->setIcon(QIcon(":/icons/javascript-warning"));
    else javaScriptActionPointer->setIcon(QIcon(":/icons/privacy-mode"));
}

void BrowserWindow::updateSearchEngineActions(const QString &searchEngine) const
{
    // Initialize the custom search engine flag.
    bool customSearchEngine = false;

    if (searchEngine == "Mojeek")  // Mojeek.
    {
        searchEngineMojeekActionPointer->setChecked(true);
    }
    else if (searchEngine == "Monocles")  // Monocles.
    {
        searchEngineMonoclesActionPointer->setChecked(true);
    }
    else if (searchEngine == "MetaGer")  // MetaGer.
    {
        searchEngineMetagerActionPointer->setChecked(true);
    }
    else if (searchEngine == "Google")  // Google.
    {
        searchEngineGoogleActionPointer->setChecked(true);
    }
    else if (searchEngine == "Bing")  // Bing.
    {
        searchEngineBingActionPointer->setChecked(true);
    }
    else if (searchEngine == "Yahoo")  // Yahoo.
    {
        searchEngineYahooActionPointer->setChecked(true);
    }
    else  // Custom search engine.
    {
        // Check the user agent.
        searchEngineCustomActionPointer->setChecked(true);

        // Set the custom search engine flag.
        customSearchEngine = true;
    }

    // Format the custom search engine.
    if (customSearchEngine)
    {
        // Enable the custom search engine.
        searchEngineCustomActionPointer->setEnabled(true);

        // Set the custom search engine text.
        searchEngineCustomActionPointer->setText(searchEngine);
    }
    else
    {
        // Disable the custom search engine.
        searchEngineCustomActionPointer->setEnabled(false);

        // Reset the custom search engine text.
        searchEngineCustomActionPointer->setText(i18nc("@action", "Custom"));
    }
}

void BrowserWindow::updateUserAgentActions(const QString &userAgent) const
{
    // Initialize the custom user agent flag.
    bool customUserAgent = false;

    // Check the indicated on-the-fly user agent.
    if (userAgent == UserAgentHelper::PRIVACY_BROWSER_USER_AGENT) userAgentPrivacyBrowserActionPointer->setChecked(true);  // Privacy Browser.
    else if (userAgent == UserAgentHelper::FIREFOX_LINUX_USER_AGENT) userAgentFirefoxLinuxActionPointer->setChecked(true);  // Firefox Linux.
    else if (userAgent == UserAgentHelper::CHROMIUM_LINUX_USER_AGENT) userAgentChromiumLinuxActionPointer->setChecked(true);  // Chromium Linux.
    else if (userAgent == UserAgentHelper::FIREFOX_WINDOWS_USER_AGENT) userAgentFirefoxWindowsActionPointer->setChecked(true);  // Firefox Windows.
    else if (userAgent == UserAgentHelper::CHROME_WINDOWS_USER_AGENT) userAgentChromeWindowsActionPointer->setChecked(true);  // Chrome Windows.
    else if (userAgent == UserAgentHelper::EDGE_WINDOWS_USER_AGENT) userAgentEdgeWindowsActionPointer->setChecked(true);  // Edge Windows.
    else if (userAgent == UserAgentHelper::SAFARI_MACOS_USER_AGENT) userAgentSafariMacosActionPointer->setChecked(true);  // Safara macOS.
    else  // Custom user agent.
    {
        // Check the user agent.
        userAgentCustomActionPointer->setChecked(true);

        // Set the custom user agent flag.
        customUserAgent = true;
    }


    // Format the custom user agent.
    if (customUserAgent)
    {
        // Enable the custom user agent.
        userAgentCustomActionPointer->setEnabled(true);

        // Set the custom user agent text.
        userAgentCustomActionPointer->setText(userAgent);
    }
    else
    {
        // Disable the custom user agent.
        userAgentCustomActionPointer->setEnabled(false);

        // Reset the custom user agent text.
        userAgentCustomActionPointer->setText(i18nc("@action", "Custom"));
    }
}

void BrowserWindow::updateZoomFactorAction(const double &zoomFactor)
{
    // Set the current zoom factor.
    currentZoomFactor = zoomFactor;

    // Update the zoom factor action text, formatting the double with 2 decimal places.
    zoomFactorActionPointer->setText(ki18nc("@action", "Zoom Factor - %1").subs(zoomFactor, 0, '0', 2).toString());
}

void BrowserWindow::updateSearchEngineLabel(const QString &searchEngineString) const
{
    // Update the search engine label.
    searchEngineLabelPointer->setText(SearchEngineHelper::getSearchUrl(searchEngineString));
}

void BrowserWindow::updateUrlLineEdit(const QString &newUrl) const
{
    // Update the URL line edit if it does not have focus.
    if (!urlLineEditPointer->hasFocus())
    {
        // Update the URL line edit.
        urlLineEditPointer->setText(newUrl);
    }
}

void BrowserWindow::updateUserAgentLabel(const QString &userAgentDatabaseName) const
{
    // Update the user agent label.
    userAgentLabelPointer->setText(UserAgentHelper::getUserAgentFromDatabaseName(userAgentDatabaseName));
}
